# -*- coding: utf-8 -*-

import os
from tinydb import TinyDB, where
from tinydb.storages import JSONStorage
from tinydb.middlewares import CachingMiddleware
from collections import namedtuple


Port = namedtuple("Port", ["name", "port", "protocol", "description"])
path = os.path.join(os.path.expanduser("~"), ".whatportis_db.json")


def get_database():
    return TinyDB(path, storage=CachingMiddleware(JSONStorage))


def database_exists():
    return os.path.isfile(path)


def merge_protocols(ports=[]):
    """
    This function merge rows having the exact same data (port, name,
    description) but a different protocol.

    :param ports: the list of ports
    :return: all ports with merged protocols
    :rtype: list
    """
    keys = {}
    for port in ports:
        key = "{}-{}-{}".format(
            port.get("description"), port.get("name"), port.get("port")
        )
        if key not in keys:
            keys[key] = port
        else:
            keys[key]["protocol"] += ", {}".format(port.get("protocol"))
    return list(keys.values())


def get_ports(port, like=False):
    """
    This function creates the SQL query depending on the specified port and
    the --like option.

    :param port: the specified port
    :param like: the --like option
    :return: all ports matching the given ``port``
    :rtype: list
    """
    where_field = "port" if port.isdigit() else "name"
    db = get_database()

    if like:
        ports = db.search(where(where_field).search(port))
    else:
        ports = db.search(where(where_field) == port)

    ports = merge_protocols(ports)
    return [Port(**port) for port in ports]  # flake8: noqa (F812)



import pytest
from click.testing import CliRunner
from tinydb.storages import JSONStorage
from tinydb.middlewares import CachingMiddleware
from tinydb import TinyDB

import whatportis.db


@pytest.fixture
def runner():
    return CliRunner()


@pytest.fixture
def create_ports(tmpdir, monkeypatch):
    def _create_ports(ports):
        def get_db():
            tmp_db = tmpdir.join("db.json")
            db = TinyDB(
                str(tmp_db),
                storage=CachingMiddleware(JSONStorage)
            )

            for port in ports:
                db.insert({
                    'name': port[0],
                    'port': port[1],
                    "description": port[2],
                    "protocol": port[3]
                })
            return db

        return monkeypatch.setattr(whatportis.db, "get_database", get_db)
    return  _create_ports



# -*- coding: utf-8 -*-

import csv
import os
import sys

import click
import requests
import simplejson as json
from tinydb import TinyDB

from whatportis.db import database_exists, get_database, get_ports
from whatportis.utils import as_table


IANA_CSV_FILE = "https://www.iana.org/assignments/service-names-port-numbers/service-names-port-numbers.csv"


def unicode_csv_reader(data):
    """
    Handle Unicode CSV data
    See: https://docs.python.org/2/library/csv.html
    """

    def utf_8_encoder(unicode_csv_data):
        for line in unicode_csv_data:
            yield line.encode("utf-8")

    csv_reader = csv.reader(utf_8_encoder(data))
    for row in csv_reader:
        yield [unicode(cell, "utf-8") for cell in row]


def csv_reader(data):
    """
    Return the correct handler according to python version.
    """
    if sys.version_info[0] == 3:
        return csv.reader(data)
    return unicode_csv_reader(data)


def populate_db():
    # Download the csv
    click.echo("Downloading {}...".format(IANA_CSV_FILE))
    resp = requests.get(IANA_CSV_FILE).content.decode("utf-8")
    data = csv_reader(resp.splitlines())

    # Populate the json
    click.echo("Populating database...")
    ports = []
    for port in data:
        ports.append(
            {
                "name": port[0] if port[0] else "---",
                "port": port[1] if port[1] else "---",
                "protocol": port[2] if port[2] else "---",
                "description": port[3] if port[3] else "---",
            }
        )

    db = get_database()
    db.purge()
    db.insert_multiple(ports)
    db.close()

    click.echo("{} ports imported.".format(len(ports)))


def update_db(ctx, param, value):
    """
    This callback downloads the official CSV file provided by IANA,
    parses it and populates the database as a JSON file.
    """
    if not value or ctx.resilient_parsing:
        return

    # Check if the file exists
    if not click.confirm("Previous database will be updated, do you want to continue?"):
        click.echo("Bye.")
        ctx.exit()

    populate_db()
    ctx.exit()


# Change the server help if flask is installed
server_help = "Start a RESTful server"
try:
    import flask

    server_help += " (format <host> <port>)."
except ImportError:
    server_help += " (requires Flask)."


@click.command()
@click.version_option()
@click.argument("PORT", required=False)
@click.option(
    "--like", is_flag=True, default=False, help="Search ports containing the pattern."
)
@click.option(
    "--json",
    "use_json",
    is_flag=True,
    default=False,
    help="Format the output result as JSON.",
)
@click.option(
    "--server", default=[None] * 2, type=click.Tuple([str, int]), help=server_help
)
@click.option(
    "--update", is_flag=True, callback=update_db, expose_value=False, is_eager=True
)
def run(port, like, use_json, server):
    """Search port names and numbers."""
    if not port and not server[0]:
        raise click.UsageError("Please specify a port")

    # Check if database exists
    if not database_exists():
        click.echo("Database not found, doing first import...")
        populate_db()

    if server[0]:
        try:
            from whatportis.server import app

            app.run(host=server[0], port=server[1])
        except ImportError:
            click.echo(
                "Dependencies are missing, please use `pip install whatportis[server]`."
            )
        return

    ports = get_ports(port, like)
    if not ports:
        click.echo("No ports found for '{0}'".format(port))
        return

    if use_json:
        print(json.dumps(ports, indent=4))
    else:
        table = as_table(ports)
        print(table)



from whatportis.db import merge_protocols


def test_merge_protocols_different_ports():
    ports = [{
        "description": "My description 1",
        "name": "MyName 1",
        "port": "1234",
        "protocol": "udp"
    }, {
        "description": "My description 2",
        "name": "MyName 2",
        "port": "5678",
        "protocol": "tcp"
    }]
    result = merge_protocols(ports)
    assert sorted(result, key=lambda p: p["name"]) == sorted(ports, key=lambda p: p["name"])


def test_merge_protocols_same_port():
    ports = [{
        "description": "My description",
        "name": "MyName",
        "port": "1234",
        "protocol": "udp"
    }, {
        "description": "My description",
        "name": "MyName",
        "port": "1234",
        "protocol": "tcp"
    }]
    result = merge_protocols(ports)
    assert result == [{
        "description": "My description",
        "name": "MyName",
        "port": "1234",
        "protocol": "udp, tcp"
    }]



# -*- coding: utf-8 -*-

from flask import Flask, request, jsonify

from whatportis.db import get_database, get_ports


app = Flask(__name__)


@app.route("/ports", methods=["GET"])
def all_ports():
    """Returns all ports"""
    return jsonify({"ports": get_database().all()})


@app.route("/ports/<pattern>", methods=["GET"])
def specific_ports(pattern):
    """Returns all ports matching the given pattern"""
    like = "like" in request.args
    return jsonify({"ports": get_ports(pattern, like)})



import json
import re

from whatportis.cli import run


def test_search_with_no_arg(runner):
    result = runner.invoke(run, [])
    regex = re.compile('^Usage:')
    assert result.exit_code == 2
    assert regex.match(result.output)


def test_search_with_multiple_args(runner):
    result = runner.invoke(run, ['foo', 'bar'])
    regex = re.compile('^Usage:')
    assert result.exit_code == 2
    assert regex.match(result.output)


def test_search_str_port(runner, create_ports):
    create_ports([["foo", "1234", "My description", "tcp"]])
    result = runner.invoke(run, ['foo'])
    assert result.exit_code == 0
    assert "foo" in result.output
    assert "1234" in result.output
    assert "My description" in result.output
    assert "tcp" in result.output


def test_search_int_port(runner, create_ports):
    create_ports([["foo", "1234", "My description", "tcp"]])
    result = runner.invoke(run, ['1234'])
    assert result.exit_code == 0
    assert "foo" in result.output
    assert "1234" in result.output
    assert "My description" in result.output
    assert "tcp" in result.output


def test_search_str_like_port(runner, create_ports):
    create_ports([
        ["foo", "1", "My description", "tcp"],
        ["foobar", "2", "My description", "udp"],
        ["baz", "3", "My description", "tcp"]
    ])

    result = runner.invoke(run, ['foo', '--like'])
    assert result.exit_code == 0
    assert result.output.count('foo') == 2

    result = runner.invoke(run, ['bar', '--like'])
    assert result.exit_code == 0
    assert result.output.count('bar') == 1


def test_search_int_like_port(runner, create_ports):
    create_ports([
        ["foo", "8", "My description", "tcp"],
        ["foobar", "9", "My description", "udp"],
        ["baz", "80", "My description", "tcp"]
    ])

    result = runner.invoke(run, ['8', '--like'])
    assert result.exit_code == 0
    assert result.output.count('8') == 2


def test_search_port_returns_json(runner, create_ports):
    create_ports([["foo", "1234", "My description", "tcp"]])
    result = runner.invoke(run, ['--json', 'foo'])

    result = json.loads(result.output)
    wanted = [{
        'description': 'My description',
        'name': 'foo',
        'port': '1234',
        'protocol': 'tcp'
    }]
    assert sorted(result, key=lambda p: p["name"]) == sorted(wanted, key=lambda p: p["name"])



from setuptools import setup

entry_points = {
    'console_scripts': [
        'whatportis=whatportis.cli:run',
    ]
}
readme = open('README.rst').read()

setup(
    name="whatportis",
    version="0.8",
    url='http://github.com/ncrocfer/whatportis',
    author='Nicolas Crocfer',
    author_email='ncrocfer@gmail.com',
    description="A command to search port names and numbers",
    long_description=readme,
    packages=['whatportis'],
    include_package_data=True,
    install_requires=[
        "simplejson",
        "tinydb",
        "requests",
        "prettytable",
        "click"
    ],
    extras_require={
        "dev": [
            "pytest",
            "tox"
        ],
        "server": [
            "flask"
        ]
    },
    entry_points=entry_points,
    classifiers=(
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Natural Language :: English',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4'
    ),
)



# -*- coding: utf-8 -*-

from prettytable import PrettyTable


def as_table(ports):
    """
    This function returns a pretty table used to display the port results.

    :param ports: list of found ports
    :return: the table to display
    """
    table = PrettyTable(["Name", "Port", "Protocol", "Description"])
    table.align["Name"] = "l"
    table.align["Description"] = "l"
    table.padding_width = 1

    for port in ports:
        table.add_row(port)

    return table
